#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

"""
OutboundMessageFactory Module.
"""


class OutboundMessageFactory(ABC):
    """Serialize messages to be sent to WolkAbout IoT Platform."""

    @abstractmethod
    def make_from_sensor_reading(self, reading):
        """
        Serialize a sensor reading to be sent to WolkAbout IoT Platform.

        :param reading: Reading to be serialized
        :type reading: wolk.models.SensorReading.SensorReading
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_alarm(self, alarm):
        """
        Serialize an alarm event to be sent to WolkAbout IoT Platform.

        :param alarm: Alarm to be serialized
        :type alarm: wolk.models.Alarm.Alarm
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_actuator_status(self, actuator):
        """
        Serialize an actuator status to be sent to WolkAbout IoT Platform.

        :param actuator: Actuator status to be serialized
        :type actuator: wolk.models.ActuatorStatus.ActuatorStatus
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_firmware_status(self, firmware_status):
        """
        Report the current status of the firmware update process.

        :param firmware_status: Current status of the firmware update process
        :type firmware_status: wolk.models.FirmwareStatus.FirmwareStatus
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_chunk_request(self, file_name, chunk_index, chunk_size):
        """
        Request a chunk of the firmware file from WolkAbout IoT Platform.

        :param file_name: Name of the file that contains the requested chunk
        :type file_name: str
        :param chunk_index: Index of the requested chunk
        :type chunk_index: int
        :param chunk_size: Size of the requested chunk
        :type chunk_size: int
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_firmware_version(self, version):
        """
        Report the current firmware version to WolkAbout IoT Platform.

        :param version: Current firmware version
        :type version: str
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_keep_alive_message(self):
        """
        Make a ping message to be sent to WolkAbout IoT Platform.

        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def make_from_configuration(self, configuration):
        """
        Serialize device's configuration to be sent to WolkAbout IoT Platform.

        :param configuration: Device's current configuration
        :type configuration: dict
        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""OSMQTTConnectivityService Module."""

from paho.mqtt import client as mqtt

from wolk.interfaces.ConnectivityService import ConnectivityService
from wolk.models.InboundMessage import InboundMessage
from wolk import LoggerFactory


MQTT_KEEP_ALIVE_INTERVAL = 90


class OSMQTTConnectivityService(ConnectivityService):
    """Handle sending and receiving MQTT messages.

    :ivar ca_cert: Path to ca.crt file used for TLS
    :vartype ca_cert: str
    :ivar client: MQTT client used to send/receive data
    :vartype client: paho.mqtt.Client
    :ivar connected: State of the connection
    :vartype connected: bool
    :ivar connected_rc: return code of the connection
    :vartype connected_rc: int
    :ivar device: Holds authentication information and actuator references
    :vartype device: wolk.models.Device.Device
    :ivar host: Address of the MQTT broker
    :vartype host: str
    :ivar inbound_message_listener: Callback function for inbound messages
    :vartype inbound_message_listener: function
    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar port: Port to which to send messages
    :vartype port: int
    :ivar qos: Quality of Service for the MQTT connection
    :vartype qos: int
    :ivar topics: List of topics to subscribe to
    :vartype topics: list
    """

    def __init__(
        self,
        device,
        topics,
        qos=2,
        host="api-demo.wolkabout.com",
        port=8883,
        ca_cert=None,
    ):
        """Provide the connection to the WolkAbout IoT Platform.

        :param device: Contains device key, password and actuator references
        :type device: wolk.models.Device.Device
        :param topics: List of topics to subscribe to
        :type topics: list
        :param qos: Quality of Service for MQTT connection
        :type qos: int or None
        :param host: Address of the MQTT broker
        :type host: str or None
        :param port: Port to which to send messages
        :type port: int or None
        :param ca_cert: Certificate file used to encrypt the connection
        :type ca_cert: str or None
        """
        self.device = device
        self.qos = qos
        self.host = host
        self.port = port
        self.connected = False
        self.connected_rc = None
        self.inbound_message_listener = None
        self.client = None
        self.topics = topics
        self.ca_cert = ca_cert
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            "Init:  Device key: %s ; Device password: %s ;"
            " Actuator references: %s ;"
            " QoS: %s ; Host: %s ; Port: %s ; CA certificate: %s ",
            self.device.key,
            self.device.password,
            self.device.actuator_references,
            self.qos,
            self.host,
            self.port,
            self.ca_cert,
        )

    def set_inbound_message_listener(self, on_inbound_message):
        """
        Set the callback function to handle inbound messages.

        :param on_inbound_message: Function that handles inbound messages
        :type on_inbound_message: function
        """
        self.logger.debug(
            "set_inbound_message_listener called - "
            "inbound_message_listener set to: %s",
            on_inbound_message,
        )
        self.inbound_message_listener = on_inbound_message

    def on_mqtt_message(self, client, userdata, message):
        """
        Serialize inbound messages and pass them to inbound message listener.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: Private user data set in Client()
        :type userdata: str
        :param message: Class with members: topic, payload, qos, retain.
        :type message: paho.mqtt.MQTTMessage
        """
        self.logger.debug("on_mqtt_message called")
        if not message:
            return

        self.inbound_message_listener(
            InboundMessage(message.topic, message.payload)
        )
        self.logger.debug(
            "Received from topic: %s ; Payload: %s",
            message.topic,
            str(message.payload),
        )

    def on_mqtt_connect(self, client, userdata, flags, rc):
        """
        Handle when the client receives a CONNACK response from the server.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param flags: Response flags sent by the broker
        :type flags: int
        :param rc: Connection result
        :type rc: int
        """
        self.logger.debug("on_mqtt_connect called with rc: %s", rc)
        if rc == 0:  # Connection successful

            self.connected = True
            self.connected_rc = 0
            # Subscribing in on_mqtt_connect() means if we lose the connection
            # and reconnect then subscriptions will be renewed.
            if self.topics:
                for topic in self.topics:
                    self.client.subscribe(topic, 2)
            self.logger.debug(
                "on_mqtt_connect - self.connected : %s", self.connected
            )
        elif rc == 1:  # Connection refused - incorrect protocol version
            self.connected_rc = 1
        elif rc == 2:  # Connection refused - invalid client identifier
            self.connected_rc = 2
        elif rc == 3:  # Connection refused - server unavailable
            self.connected_rc = 3
        elif rc == 4:  # Connection refused - bad username or password
            self.connected_rc = 4
        elif rc == 5:  # Connection refused - not authorised
            self.connected_rc = 5

    def on_mqtt_disconnect(self, client, userdata, rc):
        """
        Handle when the client disconnects from the broker.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param rc: Disconnection result
        :type rc: int

        :raises RuntimeError: Unexpected disconnection
        """
        if rc != 0:
            raise RuntimeError("Unexpected disconnection.")
        self.connected = False
        self.connected_rc = None
        self.logger.debug(
            "on_mqtt_disconnect - self.connected : %s", self.connected
        )

    def connect(self):
        """
        Establish the connection to the WolkAbout IoT platform.

        If there are actuators it will subscribe to topics
        that will contain actuator commands.
        Subscribes to firmware update related topics.
        Starts a loop to handle inbound messages.

        :raises RuntimeError: Reason for connection being refused
        """
        if self.connected:

            return

        self.logger.debug("connect started")
        self.client = mqtt.Client(
            client_id=self.device.key, clean_session=True
        )
        self.client.on_connect = self.on_mqtt_connect
        self.client.on_disconnect = self.on_mqtt_disconnect
        self.client.on_message = self.on_mqtt_message
        if self.ca_cert:
            self.client.tls_set(self.ca_cert)
            self.client.tls_insecure_set(True)
        self.client.username_pw_set(self.device.key, self.device.password)
        self.client.will_set(
            "lastwill/" + self.device.key, "Gone offline", 2, False
        )

        self.logger.debug(
            "client.connect called : host: %s ; port:%s, ca_cert:%s "
            "; username:%s ; password:%s",
            self.host,
            self.port,
            self.ca_cert,
            self.device.key,
            self.device.password,
        )
        self.client.connect(
            self.host, self.port, keepalive=MQTT_KEEP_ALIVE_INTERVAL
        )

        self.client.loop_start()

        while True:

            if self.connected_rc is None:
                continue

            if self.connected_rc == 0:
                break

            elif self.connected_rc == 1:
                raise RuntimeError(
                    "Connection refused - incorrect protocol version"
                )
                break

            elif self.connected_rc == 2:
                raise RuntimeError(
                    "Connection refused - invalid client identifier"
                )
                break

            elif self.connected_rc == 3:
                raise RuntimeError("Connection refused - server unavailable")
                break

            elif self.connected_rc == 4:
                raise RuntimeError(
                    "Connection refused - bad username or password"
                )
                break

            elif self.connected_rc == 5:
                raise RuntimeError("Connection refused - not authorised")
                break

        self.logger.debug("calling subscribe with topics: %s", self.topics)
        for topic in self.topics:
            self.client.subscribe(topic, 2)
        self.logger.debug("connect ended")

    def disconnect(self):
        """Disconnects the device from the WolkAbout IoT Platform."""
        self.logger.debug("disconnect")
        self.client.publish("lastwill/" + self.device.key, "Gone offline")
        self.client.loop_stop()
        self.client.disconnect()

    def publish(self, outbound_message):
        """
        Publish serialized data to WolkAbout IoT Platform.

        :param outbound_message: Message to be published
        :type outbound_message: wolk.models.OutboundMessage.OutboundMessage

        :returns: result
        :rtype: bool
        """
        if outbound_message is None:
            self.logger.warning("No message to publish!")
            return False

        if not self.connected:
            self.logger.warning(
                "Not connected, unable to publish. Topic: %s ; Payload: %s",
                outbound_message.topic,
                outbound_message.payload,
            )
            return False

        info = self.client.publish(
            outbound_message.topic, outbound_message.payload, self.qos
        )

        if info.rc == mqtt.MQTT_ERR_SUCCESS:

            self.logger.debug(
                "Published to topic: %s ; Payload: %s",
                outbound_message.topic,
                outbound_message.payload,
            )
            return True

        elif info.is_published():

            self.logger.debug(
                "Published to topic: %s ; Payload: %s",
                outbound_message.topic,
                outbound_message.payload,
            )
            return True

        else:

            self.logger.warning(
                "Publishing failed! Topic: %s ; Payload: %s",
                outbound_message.topic,
                outbound_message.payload,
            )
            return False



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""OSFirmwareUpdate Module."""

import base64
import hashlib
import math
from threading import Timer

from wolk.models.FirmwareErrorType import FirmwareErrorType
from wolk.models.FirmwareStatus import FirmwareStatus
from wolk.models.FirmwareStatusType import FirmwareStatusType
from wolk.models.FirmwareUpdateStateType import FirmwareUpdateStateType
from wolk.interfaces.FirmwareUpdate import FirmwareUpdate
from wolk import LoggerFactory


class OSFirmwareUpdate(FirmwareUpdate):
    """
    Responsible for everything related to the firmware update process.

    :ivar auto_install: install upon receiving the firmware file
    :vartype auto_install: bool
    :ivar expected_number_of_chunks: expected number of firmware chunks
    :vartype expected_number_of_chunks: int
    :ivar file_hash: hash of firmware file received from the platform
    :vartype file_hash: str
    :ivar file_name: firmware file name
    :vartype file_name: str
    :ivar file_size: firmware file size
    :vartype file_size: int
    :ivar firmware_handler: implementation of FirmwareHandler interface
    :vartype firmware_handler: wolk.interfaces.FirmwareHandler.FirmwareHandler
    :ivar install_timer: timer for install countdown
    :vartype install_timer: threading.Timer
    :ivar last_packet_hash: hash of last valid packet received
    :vartype last_packet_hash: str
    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar max_retries: number of retries for a firmware packet
    :vartype max_retries: int
    :ivar minimum_packet_size: minimal size of a firmware chunk
    :vartype minimum_packet_size: int
    :ivar next_chunk_index: index of next firmware chunk
    :vartype next_chunk_index: int
    :ivar on_file_packet_request_callback: callback for chunk request
    :vartype on_file_packet_request_callback: function
    :ivar on_status_callback: callback for reporting firmware update status
    :vartype on_status_callback: function
    :ivar request_timeout: countdown timer for request timeout
    :vartype request_timeout: threading.Timer
    :ivar retry_count: current retry count
    :vartype retry_count: int
    :ivar state: current firmware installation state
    :vartype state: wolk.models.FirmwareUpdateStateType.FirmwareUpdateStateType
    """

    def __init__(self, firmware_handler=None):
        """
        Responsible for firmware update flow.

        :param firmware_handler: Responsible for the firmware file itself
        :type firmware_handler: wolk.interfaces.FirmwareHandler.FirmwareHandler or None
        """
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug("Init - firmware_handler: %s", firmware_handler)
        self.firmware_handler = firmware_handler
        self.state = FirmwareUpdateStateType.IDLE
        self.max_retries = 3
        self.minimum_packet_size = 65
        self.on_status_callback = None
        self.on_file_packet_request_callback = None
        self.file_name = None
        self.file_size = None
        self.file_hash = None
        self.auto_install = None
        self.next_chunk_index = None
        self.expected_number_of_chunks = None
        self.retry_count = None
        self.request_timeout = None
        self.install_timer = None
        self.last_packet_hash = 32 * b"\x00"

        if self.firmware_handler:

            self.firmware_handler.set_url_download_result_callback(
                self.handle_url_download_result
            )

    def set_on_file_packet_request_callback(self, on_file_packet_request_callback):
        """
        Set the callback function for requesting file packets.

        :param on_file_packet_request_callback: Function to be called
        :type on_file_packet_request_callback: function
        """
        self.logger.debug(
            "set_on_file_packet_request_callback called - "
            "on_file_packet_request_callback = %s",
            on_file_packet_request_callback,
        )
        self.on_file_packet_request_callback = on_file_packet_request_callback

    def set_on_status_callback(self, on_status_callback):
        """
        Set the callback function for reporting firmware status.

        :param on_status_callback: Function to be called
        :type on_status_callback: function
        """
        self.logger.debug(
            "set_on_status_callback called - " "on_status_callback = %s",
            on_status_callback,
        )
        self.on_status_callback = on_status_callback

    def handle_url_download_result(self, result):
        """
        Receive the result of URL download from the firmware handler.

        :param result: Result of the URL download
        :type result: bool
        """
        self.logger.debug("handle_url_download_result called - Result: %s", result)
        if not self.firmware_handler:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.logger.debug(
                "handle_url_download_result - No firmware_handler, "
                "reporting DFU disabled"
            )
            self.on_status_callback(status)
            return

        if result:

            self.state = FirmwareUpdateStateType.FILE_OBTAINED
            status = FirmwareStatus(FirmwareStatusType.FILE_READY)
            self.logger.debug(
                "handle_url_download_result - result valid, " "reporting file ready"
            )
            self.on_status_callback(status)

            if self.auto_install:

                self.logger.debug(
                    "handle_url_download_result - auto_install is true, "
                    "handling install"
                )
                self.handle_install()

        else:

            self.logger.debug(
                "handle_url_download_result - result invalid, " "abort and report error"
            )
            self.firmware_handler.update_abort()
            self.reset_state()
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSPECIFIED_ERROR
            )
            self.on_status_callback(status)

    def handle_file_upload(self, firmware_command):
        """
        Handle the file upload command received from the platform.

        :param firmware_command: Firmware command received
        :type firmware_command: wolk.models.FirmwareCommand.FirmwareCommand
        """
        self.logger.info(
            "Received firmware command - Command: FILE_UPLOAD ; "
            "File name: %s ; File size: %s ; "
            "File hash: %s ; Auto install: %s",
            firmware_command.file_name,
            firmware_command.file_size,
            firmware_command.file_hash,
            firmware_command.auto_install,
        )
        if not self.firmware_handler:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.logger.error("No firmware_handler, reporting DFU disabled")
            self.on_status_callback(status)
            return

        if self.state != FirmwareUpdateStateType.IDLE:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSPECIFIED_ERROR
            )
            self.logger.error("State not idle, reporting unspecified error")
            self.state = FirmwareUpdateStateType.IDLE
            self.on_status_callback(status)
            return

        if (
            self.firmware_handler.max_file_size == 0
            or self.firmware_handler.chunk_size == 0
        ):

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.logger.error("firmware_handler not configured, reporting DFU disabled")
            self.on_status_callback(status)
            return

        if self.firmware_handler.max_file_size < firmware_command.file_size:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSUPPORTED_FILE_SIZE
            )
            self.logger.error("File size not supported, reporting size error")
            self.on_status_callback(status)
            return

        if not self.firmware_handler.update_start(
            firmware_command.file_name, firmware_command.file_size
        ):

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSPECIFIED_ERROR
            )
            self.logger.error(
                "Failed to create temporary file, reporting unspecified error"
            )
            self.on_status_callback(status)
            return

        self.file_name = firmware_command.file_name
        self.file_size = firmware_command.file_size
        self.file_hash = firmware_command.file_hash
        self.auto_install = firmware_command.auto_install
        self.expected_number_of_chunks = math.ceil(
            self.file_size / self.firmware_handler.chunk_size
        )
        self.next_chunk_index = 0
        self.retry_count = 0
        self.state = FirmwareUpdateStateType.FILE_TRANSFER

        status = FirmwareStatus(FirmwareStatusType.FILE_TRANSFER)
        self.logger.info("Initializing file transfer and requesting first chunk...")
        self.on_status_callback(status)

        if firmware_command.file_size < self.firmware_handler.chunk_size:
            self.on_file_packet_request_callback(
                self.file_name, self.next_chunk_index, firmware_command.file_size + 64
            )
        else:
            self.on_file_packet_request_callback(
                self.file_name, self.next_chunk_index, self.firmware_handler.chunk_size + 64
            )

        self.request_timeout = Timer(60.0, self.handle_abort)
        self.request_timeout.start()

    def handle_url_download(self, firmware_command):
        """
        Handle the URL download command received from the platform.

        :param firmware_command: Firmware command received
        :type firmware_command: wolk.models.FirmwareCommand.FirmwareCommand
        """
        self.logger.info(
            "Received URL download command; File URL: %s ; Auto install:%s",
            firmware_command.file_url,
            firmware_command.auto_install,
        )
        if not self.firmware_handler:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.logger.error("No firmware_handler, reporting DFU disabled")
            self.on_status_callback(status)
            return

        if self.state != FirmwareUpdateStateType.IDLE:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSPECIFIED_ERROR
            )
            self.logger.error("State not idle, reporting unspecified error")
            self.state = FirmwareUpdateStateType.IDLE
            self.on_status_callback(status)
            return

        if not self.firmware_handler.update_start_url_download(
            firmware_command.file_url
        ):

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.MALFORMED_URL
            )
            self.logger.error("Malformed URL!")
            self.on_status_callback(status)
            return

        self.auto_install = firmware_command.auto_install
        self.state = FirmwareUpdateStateType.URL_DOWNLOAD
        status = FirmwareStatus(FirmwareStatusType.FILE_TRANSFER)
        self.logger.info("Initializing URL file transfer...")
        self.on_status_callback(status)

    def handle_install(self):
        """Handle the install command received from the platform."""
        self.logger.debug("handle_install called")
        if not self.firmware_handler:

            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.logger.error("No firmware_handler, reporting DFU disabled")
            self.on_status_callback(status)
            return

        if self.state == FirmwareUpdateStateType.FILE_OBTAINED:

            if not self.firmware_handler.persist_version(self.firmware_handler.version):

                self.logger.error(
                    "Unable to persist version, " "aborting firmware update process"
                )
                self.firmware_handler.update_abort()
                self.reset_state()
                status = FirmwareStatus(
                    FirmwareStatusType.ERROR, FirmwareErrorType.INSTALLATION_FAILED
                )
                self.on_status_callback(status)
                return

            self.state = FirmwareUpdateStateType.INSTALL
            status = FirmwareStatus(FirmwareStatusType.INSTALLATION)
            self.logger.info("Beginning firmware installation process")
            self.on_status_callback(status)

            self.install_timer = Timer(
                5.0, self.firmware_handler.update_finalize
            )  # For possible abort command
            self.install_timer.start()

        elif self.state == FirmwareUpdateStateType.INSTALL:

            self.logger.debug(
                "handle_install - Ignore install command, "
                "installation already started"
            )

        else:  # Returns an error in all other states

            self.logger.error("Invalid flow, aborting and reporting unspecified error")
            self.firmware_handler.update_abort()
            self.reset_state()
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSPECIFIED_ERROR
            )
            self.on_status_callback(status)

    def handle_abort(self):
        """Handle the abort command received from the platform."""
        self.logger.debug("handle_abort called")
        if self.install_timer:

            self.logger.debug("handle_abort - Stopping installation timer")
            self.install_timer.cancel()

        if not self.firmware_handler:

            self.logger.debug(
                "handle_abort - No firmware_handler, reporting DFU disabled"
            )
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.on_status_callback(status)
            return

        if self.state != FirmwareUpdateStateType.IDLE:

            self.logger.info("Received abort command from platform")
            self.firmware_handler.update_abort()
            self.reset_state()
            status = FirmwareStatus(FirmwareStatusType.ABORTED)
            self.on_status_callback(status)

        else:

            self.logger.error("Invalid flow, reporting unspecified error")
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.UNSPECIFIED_ERROR
            )
            self.on_status_callback(status)

    def handle_packet(self, packet):
        """
        Handle the firmware file chunk packet received from the platform.

        :param packet: Firmware file chunk received
        :type packet: wolk.models.FileTransferPacket.FileTransferPacket
        """
        self.logger.debug(
            "handle_packet called - Previous hash: %s ; "
            "Data size: %s ; Current hash: %s",
            packet.previous_hash,
            len(packet.data),
            packet.current_hash,
        )
        if self.state != FirmwareUpdateStateType.FILE_TRANSFER:

            self.logger.debug(
                "handle_packet - State not file transfer, ignoring packet"
            )
            return

        if self.request_timeout:
            self.logger.debug("handle_packet - Canceling request timeout timer")
            self.request_timeout.cancel()

        if not self.firmware_handler:

            self.logger.error("No firmware_handler, reporting DFU disabled")
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.on_status_callback(status)
            return

        if not self.validate_packet(packet):

            self.logger.debug("handle_packet - Packet not valid")
            self.retry_count += 1

            if self.retry_count >= self.max_retries:

                self.logger.error(
                    "Retry count exceeded, aborting firmware update process"
                )
                self.firmware_handler.update_abort()
                self.reset_state()
                status = FirmwareStatus(
                    FirmwareStatusType.ERROR, FirmwareErrorType.RETRY_COUNT_EXCEEDED
                )
                self.on_status_callback(status)
                return

            else:

                self.logger.info("Requesting chunk #%s again", self.next_chunk_index)
                self.on_file_packet_request_callback(
                    self.file_name,
                    self.next_chunk_index,
                    self.firmware_handler.chunk_size + 64,
                )

                self.request_timeout = Timer(60.0, self.handle_abort)
                self.request_timeout.start()
                return

        if self.last_packet_hash != packet.previous_hash:

            self.logger.error("Received chunk out of order!")
            self.retry_count += 1

            if self.retry_count >= self.max_retries:

                self.logger.error(
                    "Retry count exceeded, aborting firmware update process"
                )
                self.firmware_handler.update_abort()
                self.reset_state()
                status = FirmwareStatus(
                    FirmwareStatusType.ERROR, FirmwareErrorType.RETRY_COUNT_EXCEEDED
                )
                self.on_status_callback(status)
                return

            else:

                self.logger.info("Requesting chunk #%s again", self.next_chunk_index)
                self.on_file_packet_request_callback(
                    self.file_name,
                    self.next_chunk_index,
                    self.firmware_handler.chunk_size + 64,
                )

                self.request_timeout = Timer(60.0, self.handle_abort)
                self.request_timeout.start()
                return

        self.last_packet_hash = packet.current_hash

        if not self.firmware_handler.write_chunk(packet.data):

            self.logger.error(
                "Failed to write chunk, aborting firmware update process!"
            )
            self.firmware_handler.update_abort()
            self.reset_state()
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_SYSTEM_ERROR
            )
            self.on_status_callback(status)
            return

        self.next_chunk_index += 1

        if self.next_chunk_index < self.expected_number_of_chunks:

            self.logger.debug(
                "handle_packet - Wrote chunk successfully, " "requesting next chunk"
            )
            self.on_file_packet_request_callback(
                self.file_name,
                self.next_chunk_index,
                self.firmware_handler.chunk_size + 64,
            )

            self.request_timeout = Timer(60.0, self.handle_abort)
            self.request_timeout.start()
            return

        if not self.validate_firmware_file():

            self.logger.error(
                "Firmware file not valid, aborting firmware update process!"
            )
            self.firmware_handler.update_abort()
            self.reset_state()
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_SYSTEM_ERROR
            )
            self.on_status_callback(status)
            return

        self.state = FirmwareUpdateStateType.FILE_OBTAINED
        self.request_timeout.cancel()
        self.request_timeout = None
        self.logger.info("Firmware file validated, ready for installation")
        status = FirmwareStatus(FirmwareStatusType.FILE_READY)
        self.on_status_callback(status)

        if self.auto_install:

            self.logger.info("Auto install is enabled, beginning installation..")
            self.handle_install()

    def report_result(self):
        """Report the result of the firmware installation process."""
        self.logger.debug("report_result called")
        if self.firmware_handler is not None:

            persisted_version = self.firmware_handler.unpersist_version()
            if not persisted_version:
                self.logger.debug("report_result - No persisted version")
                return

            if persisted_version == self.firmware_handler.version:

                self.logger.info(
                    "Firmware version unchanged, reporting installation failed"
                )
                status = FirmwareStatus(
                    FirmwareStatusType.ERROR, FirmwareErrorType.INSTALLATION_FAILED
                )
                self.on_status_callback(status)
                return

            self.logger.info(
                "Firmware version changed, reporting installation completed"
            )
            status = FirmwareStatus(FirmwareStatusType.COMPLETED)
            self.on_status_callback(status)

    def reset_state(self):
        """Reset the state of the firmware update handler."""
        self.logger.debug("reset_state called")
        self.state = FirmwareUpdateStateType.IDLE
        self.file_name = None
        self.file_size = None
        self.file_hash = None
        self.auto_install = None
        self.next_chunk_index = None
        self.expected_number_of_chunks = None
        self.retry_count = None
        self.request_timeout = None
        self.install_timer = None
        self.last_packet_hash = 32 * b"\x00"

    def validate_packet(self, packet):
        """
        Validate received firmware file packet.

        Checks if the packed is bigger than the minimum packet size
        and checks if the sha256 hash of the data matches the hash
        received from the platform

        :param packet: Packet to validate
        :type packet: wolk.models.FileTransferPacket.FileTransferPacket

        :returns: valid
        :rtype: bool
        """
        self.logger.debug("validate_packet called")
        if (
            len(packet.previous_hash) + len(packet.data) + len(packet.current_hash)
            < self.minimum_packet_size
        ):

            self.logger.error("Received packet size too small!")
            return False

        else:

            data_hash = hashlib.sha256(packet.data).digest()

            if packet.current_hash != data_hash:

                self.logger.error(
                    "Data hash '%s' doesn't match expected hash '%s'!",
                    data_hash,
                    packet.current_hash,
                )
                return False

            else:

                self.logger.debug(
                    "validate_packet - Data hash matches current_hash,"
                    " returning True"
                )
                return True

    def validate_firmware_file(self):
        """
        Validate the completed firmware file.

        Compares the sha256 hash of the received file with the file hash
        received from the platform.

        :returns: valid
        :rtype: bool
        """
        self.logger.debug("validate_firmware_file called")
        if not self.firmware_handler:

            self.logger.error("No firmware_handler, reporting DFU disabled")
            status = FirmwareStatus(
                FirmwareStatusType.ERROR, FirmwareErrorType.FILE_UPLOAD_DISABLED
            )
            self.on_status_callback(status)
            return

        sha256_command_file_hash = base64.b64decode(
            self.file_hash + ("=" * (-len(self.file_hash) % 4))
        )
        sha256_file_hash = hashlib.sha256()

        for x in range(0, self.expected_number_of_chunks):

            chunk = self.firmware_handler.read_chunk(x)
            if not chunk:
                self.logger.error("File size too small!")
                return False

            sha256_file_hash.update(chunk)

        sha256_file_hash = sha256_file_hash.digest()

        valid = sha256_command_file_hash == sha256_file_hash
        self.logger.debug("validate_firmware_file - returning valid: %s", valid)
        return valid



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""OSOutboundMessageQueue Module."""

from collections import deque

from wolk.interfaces.OutboundMessageQueue import OutboundMessageQueue
from wolk import LoggerFactory


class OSOutboundMessageQueue(OutboundMessageQueue):
    """
    Store messages before they are sent to the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar queue: Double ended queue used to store messages
    :vartype queue: collections.deque
    """

    def __init__(self):
        """Create a double ended queue to store messages."""
        self.queue = deque()
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug("Init")

    def put(self, message):
        """
        Add the message to the queue.

        Combines sensor reading messages of the same reference
        into a single message.

        :param message: Message to place in the queue
        :type message: wolk.models.OutboundMessage.OutboundMessage
        """
        if not message:
            return

        if "reading" not in message.topic:
            self.queue.append(message)
            self.logger.debug(
                "put - Queue size: %s ; Topic: %s ; Payload: %s",
                len(self.queue),
                message.topic,
                message.payload,
            )
            return

        reading_reference = message.topic.split("/")[-1]

        present_in_queue = False

        for stored_message in self.queue:
            if "reading" not in message.topic:
                continue
            if reading_reference == stored_message.topic.split("/")[-1]:
                present_in_queue = True
                break

        if not present_in_queue:
            self.queue.append(message)
            self.logger.debug(
                "put - Queue size: %s ; Topic: %s ; Payload: %s",
                len(self.queue),
                message.topic,
                message.payload,
            )
            return

        readings = 0
        max_data = 1

        for stored_message in self.queue:
            if "reading" not in stored_message.topic:
                continue
            if reading_reference == stored_message.topic.split("/")[-1]:
                readings += 1
                data_count = stored_message.payload.count('"data"')
                if data_count > max_data:
                    max_data = data_count

        if readings > 0 and max_data == 1:
            for stored_message in self.queue:
                if reading_reference == stored_message.topic.split("/")[-1]:
                    stored_message.payload = "[" + stored_message.payload
                    stored_message.payload += "," + message.payload + "]"
                    self.logger.debug(
                        "put - Queue size: %s ; Topic: %s ; Payload: %s",
                        len(self.queue),
                        message.topic,
                        message.payload,
                    )
                    break

        if max_data > 1:
            for stored_message in self.queue:
                if "reading" not in stored_message.topic:
                    continue
                if reading_reference == stored_message.topic.split("/")[-1]:
                    data_count = stored_message.payload.count('"data"')
                    if max_data > data_count:
                        continue

                    stored_message.payload = stored_message.payload[:-1]
                    stored_message.payload += "," + message.payload + "]"
                    self.logger.debug(
                        "put - Queue size: %s ; Topic: %s ; Payload: %s",
                        len(self.queue),
                        message.topic,
                        message.payload,
                    )

    def get(self):
        """
        Take the first message from the queue.

        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage or None
        """
        if len(self.queue) == 0:
            return None

        message = self.queue.popleft()
        self.logger.debug(
            "get - Queue size: %s ; Topic: %s ; Payload: %s",
            len(self.queue),
            message.topic,
            message.payload,
        )
        return message

    def peek(self):
        """
        Return the first message from the queue without removing it.

        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage or None
        """
        if len(self.queue) == 0:

            self.logger.debug("peek - Queue size: %s", len(self.queue))
            return

        else:
            message = self.queue[0]
            self.logger.debug(
                "peek - Queue size: %s ; Topic: %s ; Payload: %s",
                len(self.queue),
                message.topic,
                message.payload,
            )
            return message



from .ActuationHandler import ActuationHandler
from .ActuatorStatusProvider import ActuatorStatusProvider
from .ConfigurationHandler import ConfigurationHandler
from .ConfigurationProvider import ConfigurationProvider
from .ConnectivityService import ConnectivityService
from .FirmwareInstaller import FirmwareInstaller
from .FirmwareURLDownloadHandler import FirmwareURLDownloadHandler
from .InboundMessageDeserializer import InboundMessageDeserializer
from .OutboundMessageFactory import OutboundMessageFactory
from .OutboundMessageQueue import OutboundMessageQueue


__all__ = [
    "ActuationHandler",
    "ActuatorStatusProvider",
    "ConfigurationHandler",
    "ConfigurationProvider",
    "ConnectivityService",
    "FirmwareInstaller",
    "FirmwareURLDownloadHandler",
    "InboundMessageDeserializer",
    "OutboundMessageFactory",
    "OutboundMessageQueue",
]



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

"""
ConfigurationProvider Module.
"""


class ConfigurationProvider(ABC):
    """Read the device's current configuration options."""

    @abstractmethod
    def get_configuration(self):
        """
        Get current configuration options.

        Reads device configuration and returns it as a dictionary
        with device configuration reference as key,
        and device configuration value as value.

        Must be implemented as non blocking.
        Must be implemented as thread safe.

        :returns: configuration
        :rtype: dict
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""FileSystemFirmwareHandler Module."""

import os
import shutil
import tempfile
from threading import Timer
from urllib.parse import urlparse

from persistent_queue import PersistentQueue

from wolk import LoggerFactory
from wolk.interfaces.FirmwareHandler import FirmwareHandler


class FileSystemFirmwareHandler(FirmwareHandler):
    """
    Firmware handler that uses OS provided disk storage for firmware files.

    :ivar chunk_size: Desired chunk size in bytes
    :vartype chunk_size: int
    :ivar download_location: Where to store the completed firmware file
    :vartype download_location: str
    :ivar file_name: Name of firmware file
    :vartype file_name: str
    :ivar file_size: Size of firmware file in bytes
    :vartype file_size: int
    :ivar file_url: URL where there firmware file is located
    :vartype file_url: str
    :ivar firmware_installer: Installer of firmware file
    :vartype firmware_installer: wolk.interfaces.FirmwareInstaller.FirmwareInstaller
    :ivar firmware_url_download_handler: URL downloader
    :vartype firmware_url_download_handler: wolk.interfaces.FirmwareURLDownloadHandler.FirmwareURLDownloadHandler or None
    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar max_file_size: Maximum file size supported by device in bytes
    :vartype max_file_size: int
    :ivar report_result_callback: Callback for reporting URL download result
    :vartype report_result_callback: function
    :ivar temp_file: Handle of temp file used to store incomplete firmware file
    :vartype temp_file: file object
    :ivar version: Current version of the device firmware
    :vartype version: str
    :ivar version_persister: Means of storing current version on disk
    :vartype version_persister: persistent_queue.PersistentQueue
    """

    def __init__(
        self,
        version,
        chunk_size,
        max_file_size,
        firmware_installer,
        download_location,
        firmware_url_download_handler=None,
    ):
        """
        Handle file manipulation on disk storage.

        :param version: Current version of the device firmware
        :type version: str
        :param chunk_size: Desired chunk size in bytes
        :type chunk_size: int
        :param max_file_size: Maximum file size supported by device in bytes
        :type max_file_size: int
        :param firmware_installer: Installer of firmware file
        :type firmware_installer: wolk.interfaces.FirmwareInstaller.FirmwareInstaller
        :param download_location: Where to store the completed firmware file
        :type download_location: str
        :param firmware_url_download_handler: Optional URL downloader
        :type firmware_url_download_handler: wolk.interfaces.FirmwareURLDownloadHandler.FirmwareURLDownloadHandler or None
        """
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            "Init:  Version: %s ; Chunk size: %s ; Max file size: %s ; "
            "Firmware installer: %s ; Download location: '%s'  "
            "Firmware URL download handler: %s",
            version,
            chunk_size,
            max_file_size,
            firmware_installer,
            download_location,
            firmware_url_download_handler,
        )
        self.version = version
        self.chunk_size = chunk_size
        self.max_file_size = max_file_size
        self.download_location = download_location
        self.firmware_installer = firmware_installer
        self.firmware_url_download_handler = firmware_url_download_handler
        self.temp_file = None
        self.file_name = None
        self.file_size = None
        self.file_url = None
        self.report_result_callback = None
        self.version_persister = PersistentQueue("persisted_version")

    def update_start(self, file_name, file_size):
        """
        Start receiving the firmware file.

        :param file_name: Name of the firmware file
        :type file_name: str
        :param file_size: Size of the firmware file in bytes
        :type file_size: int

        :returns: result
        :rtype: bool
        """
        self.logger.debug(
            "update_start called - File name: %s ; File size: %s", file_name, file_size
        )
        if file_size > self.max_file_size:
            self.logger.debug("update_start - File size too big, returning False")
            return False

        self.temp_file = tempfile.NamedTemporaryFile(mode="a+b", delete=False)
        self.file_name = file_name
        self.file_size = file_size
        self.logger.debug("update_start - Temporary file created, returning True")
        return True

    def update_finalize(self):
        """
        Finalize firmware installation process.

        Copies the content of the temporary file to
        the desired download location.
        Calls the provided firmware installer's install_firmware function
        """
        self.logger.debug("update_finalize called")
        self.logger.debug(
            "download location: %s ;file_name: %s ;temp_file: %s",
            self.download_location,
            self.file_name,
            self.temp_file,
        )

        if not os.path.exists(os.path.abspath(self.download_location)):
            os.makedirs(os.path.abspath(self.download_location))

        firmware_file_path = os.path.join(
            os.path.abspath(self.download_location), self.file_name
        )

        if self.temp_file:
            shutil.copy2(os.path.realpath(self.temp_file.name), firmware_file_path)
            self.temp_file.close()

        self.logger.info(
            "Firmware file copied to download location from "
            "temporary file, calling firmware_installer.install_firmware "
            "with path: %s",
            firmware_file_path,
        )
        self.firmware_installer.install_firmware(firmware_file_path)

    def update_abort(self):
        """Abort the firmware update process."""
        self.logger.debug("update_abort called")

        if self.temp_file:
            self.temp_file.close()

        self.temp_file = None
        self.file_name = None
        self.file_size = None
        self.file_url = None

    def write_chunk(self, chunk):
        """
        Write a chunk of the firmware file to the temporary file.

        :param chunk: A piece of the firmware file
        :type chunk: bytes

        :returns: result
        :rtype: bool
        """
        self.logger.debug("write_chunk called - Chunk size: %s", len(chunk))

        self.temp_file.write(chunk)
        self.temp_file.flush()
        os.fsync(self.temp_file)
        self.logger.debug("write_chunk - Chunk written, returning True")
        return True

    def read_chunk(self, index):
        """
        Read a chunk of the temporary firmware file.

        :param index: Offset from the beginning of the file
        :type index: int

        :returns: chunk
        :rtype: bytes
        """
        self.logger.debug("read_chunk called - Index: %s", index)

        self.temp_file.seek(index * self.chunk_size)
        chunk = self.temp_file.read(self.chunk_size)
        self.logger.debug("read_chunk - Chunk size: %s", len(chunk))
        return chunk

    def persist_version(self, version):
        """
        Place the current firmware version into persistent storage.

        Later to be used to determine the result
        of the firmware update process

        :param version: Current firmware version
        :type version: str

        :returns: result
        :rtype: bool
        """
        self.logger.debug("persist_version called - Version: %s", version)

        self.version_persister.clear()
        self.version_persister.flush()
        self.version_persister.push(version)
        self.logger.debug("persist_version - Persisted version, returning True")
        return True

    def unpersist_version(self):
        """
        Remove the firmware version from persistent storage.

        :returns: version
        :rtype: str or None
        """
        self.logger.debug("unpersist_version called")

        if not self.version_persister.peek():
            self.version_persister.clear()
            self.logger.debug(
                "unpersist_version - No persisted version, returning None"
            )
            return None
        else:
            version = self.version_persister.pop()
            self.version_persister.flush()
            self.logger.debug(
                "unpersist_version - Persisted version found, returning %s", version
            )
            return version

    def set_url_download_result_callback(self, callback):
        """
        Set callback for reporting URL download result.

        :param callback: Function to call
        :type callback: function
        """
        self.logger.debug(
            "set_url_download_result_callback called - Callback: %s", callback
        )
        self.report_result_callback = callback

    def update_start_url_download(self, file_url):
        """
        Start firmware file URL download process.

        Calls download function from firmware_url_download_handler.
        Returns the validity of the URL and calls download function if valid.

        :param file_url: URL that contains the firmware file
        :type file_url: str

        :returns: result
        :rtype: bool
        """
        self.logger.debug("update_start_url_download called - File URL: %s", file_url)

        if not self.firmware_url_download_handler:
            self.logger.debug(
                "update_start_url_download - No firmware_url_download_handler,"
                "returning False"
            )
            return False

        parsed_url = urlparse(file_url)

        if bool(parsed_url.scheme):

            self.file_url = file_url
            self.file_name = self.file_url.split("/")[-1]
            t = Timer(
                2.0,
                self.firmware_url_download_handler.download,
                [self.file_url, self.file_name, self.report_result_callback],
            )
            t.start()
            self.logger.debug(
                "update_start_url_download - Valid URL, calling "
                "firmware_url_download_handler.download and returning True"
            )
            return True

        else:

            self.logger.debug(
                "update_start_url_download - Invalid URL, returning False"
            )
            return False



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

"""
ActuatorStatusProvider Module.
"""


class ActuatorStatusProvider(ABC):
    """Read the status of device's actuators."""

    @abstractmethod
    def get_actuator_status(self, reference):
        """
        Get current actuator status.

        Reads the status of actuator from the device
        and returns it as a tuple containing the ActuatorState and current value.
        Must be implemented as non blocking.
        Must be implemented as thread safe.

        :param reference: Actuator reference
        :type reference: str
        :returns: (state, value)
        :rtype: (wolk.models.ActuatorState.ActuatorState, int or float or str)
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
WolkConnect Module.
"""
import os

from wolk.OSOutboundMessageQueue import OSOutboundMessageQueue
from wolk.OSMQTTConnectivityService import OSMQTTConnectivityService
from wolk.JsonSingleOutboundMessageFactory import (
    JsonSingleOutboundMessageFactory,
)
from wolk.JsonSingleInboundMessageDeserializer import (
    JsonSingleInboundMessageDeserializer,
)
from wolk.JsonProtocolOutboundMessageFactory import (
    JsonProtocolOutboundMessageFactory,
)
from wolk.JsonProtocolInboundMessageDeserializer import (
    JsonProtocolInboundMessageDeserializer,
)
from wolk.OSFirmwareUpdate import OSFirmwareUpdate
from wolk.OSKeepAliveService import OSKeepAliveService
from wolk.LoggerFactory import logger_factory
from wolk.models.SensorReading import SensorReading
from wolk.models.Alarm import Alarm
from wolk.models.ActuatorCommandType import ActuatorCommandType
from wolk.models.ActuatorStatus import ActuatorStatus
from wolk.models.ConfigurationCommandType import ConfigurationCommandType
from wolk.models.FirmwareCommandType import FirmwareCommandType
from wolk.models.FirmwareStatus import FirmwareStatus
from wolk.models.FirmwareStatusType import FirmwareStatusType
from wolk.models.FirmwareErrorType import FirmwareErrorType
from wolk.models.Protocol import Protocol
from wolk.interfaces.ActuationHandler import ActuationHandler
from wolk.interfaces.ActuatorStatusProvider import ActuatorStatusProvider
from wolk.interfaces.ConfigurationHandler import ConfigurationHandler
from wolk.interfaces.ConfigurationProvider import ConfigurationProvider
from wolk.interfaces.OutboundMessageFactory import OutboundMessageFactory
from wolk.interfaces.OutboundMessageQueue import OutboundMessageQueue
from wolk.interfaces.InboundMessageDeserializer import (
    InboundMessageDeserializer,
)
from wolk.interfaces.ConnectivityService import ConnectivityService


PING_KEEP_ALIVE_INTERVAL = 60


class WolkConnect:
    """
    Exchange data with WolkAbout IoT Platform.

    :ivar connectivity_service: Means of sending/receiving data
    :vartype connectivity_service: wolk.OSMQTTConnectivityService.OSMQTTConnectivityService
    :ivar inbound_message_deserializer: Deserializer of inbound messages
    :vartype inbound_message_deserializer: wolk.JsonSingleInboundMessageDeserializer.JsonSingleInboundMessageDeserializer
    :ivar device: Contains device key and password, and actuator references
    :vartype device: wolk.models.Device.Device
    :ivar firmware_update: Firmware update handler
    :vartype firmware_update: wolk.OSFirmwareUpdate.OSFirmwareUpdate
    :ivar keep_alive_service: Keep device connected when not sending data
    :vartype keep_alive_service: wolk.OSKeepAliveService.OSKeepAliveService
    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar outbound_message_factory: Create messages to send
    :vartype outbound_message_factory: wolk.JsonSingleOutboundMessageFactory.JsonSingleOutboundMessageFactory
    :ivar outbound_message_queue: Store data before sending
    :vartype outbound_message_queue: wolk.interfaces.OutboundMessageQueue.OutboundMessageQueue
    """

    def __init__(
        self,
        device,
        protocol=Protocol.JSON_SINGLE,
        actuation_handler=None,
        actuator_status_provider=None,
        configuration_handler=None,
        configuration_provider=None,
        outbound_message_queue=None,
        keep_alive_enabled=True,
        firmware_handler=None,
        host=None,
        port=None,
        ca_cert=None,
        outbound_message_factory=None,
        inbound_message_deserializer=None,
        connectivity_service=None,
    ):
        """
        Provide communication with WolkAbout IoT Platform.

        :param device: Contains key and password, and actuator references
        :type device:  wolk.models.Device.Device
        :param actuation_handler: Handle actuation commands
        :type actuation_handler: wolk.interfaces.ActuationHandler.ActuationHandler or None
        :param actuator_status_provider: Read actuator status
        :type actuator_status_provider:  wolk.interfaces.ActuatorStatusProvider.ActuatorStatusProvider or None
        :param outbound_message_queue: Store messages before sending
        :type outbound_message_queue: wolk.interfaces.OutboundMessageQueue.OutboundMessageQueue or None
        :param keep_alive_enabled: Enable or disable the keep alive service
        :type keep_alive_enabled: bool
        :param firmware_handler: Firmware update support
        :type firmware_handler: wolk.FileSystemFirmwareHandler.FileSystemFirmwareHandler or None
        :param host: Host name or IP address of the remote broker
        :type host: str or None
        :param port: Network port of the server host to connect to
        :type port: int or None
        :param ca_cert: String path to Certificate Authority certificate file
        :type ca_cert: str or None
        :param outbound_message_factory: Creator of messages to be sent
        :type outbound_message_factory: wolk.interfaces.OutboundMessageFactory.OutboundMessageFactory or None
        :param inbound_message_deserializer: Deserializer of messages from the Platform
        :type inbound_message_deserializer: wolk.interfaces.InboundMessageDeserializer.InboundMessageDeserializer or None
        :param connectivity_service: Provider of connection to Platform
        :type connectivity_service: wolk.interfaces.ConnectivityService.ConnectivityService or None
        """
        self.logger = logger_factory.get_logger(str(self.__class__.__name__))
        self.logger.debug(
            "Init:  Device: %s ; Actuation handler: %s ; "
            "Actuator status provider: %s ; "
            "Configuration handler: %s ; Configuration provider: %s ; "
            "Outbound message queue: %s ; Keep alive enabled: %s ;"
            " Firmware handler: %s"
            "Host: %s ; Port: %s ; ca_cert: %s "
            "Outbound message factory: %s ; "
            "Inbound message deserializer: %s ; "
            "Connectivity Service: %s",
            device,
            actuation_handler,
            actuator_status_provider,
            configuration_handler,
            configuration_provider,
            outbound_message_queue,
            keep_alive_enabled,
            firmware_handler,
            host,
            port,
            ca_cert,
            outbound_message_factory,
            inbound_message_deserializer,
            connectivity_service,
        )

        self.device = device

        if actuation_handler is None:
            self.actuation_handler = None
        else:
            if not isinstance(actuation_handler, ActuationHandler):
                raise RuntimeError("Invalid actuation handler provided")
            else:
                self.actuation_handler = actuation_handler

        if actuator_status_provider is None:
            self.actuator_status_provider = None
        else:
            if not isinstance(
                actuator_status_provider, ActuatorStatusProvider
            ):
                raise RuntimeError("Invalid actuator status provider provided")
            else:
                self.actuator_status_provider = actuator_status_provider

        if configuration_handler is None:
            self.configuration_handler = None
        else:
            if not isinstance(configuration_handler, ConfigurationHandler):
                raise RuntimeError("Invalid configuration handler provided")
            else:
                self.configuration_handler = configuration_handler

        if configuration_provider is None:
            self.configuration_provider = None
        else:
            if not isinstance(configuration_provider, ConfigurationProvider):
                raise RuntimeError("Invalid configuration provider provided")
            else:
                self.configuration_provider = configuration_provider

        if outbound_message_queue is None:
            self.outbound_message_queue = OSOutboundMessageQueue()
        else:
            if not isinstance(outbound_message_queue, OutboundMessageQueue):
                raise RuntimeError("Invalid outbound message queue provided")
            else:
                self.outbound_message_queue = outbound_message_queue

        if (
            outbound_message_factory is None
            and inbound_message_deserializer is not None
        ) or (
            outbound_message_factory is not None
            and inbound_message_deserializer is None
        ):
            raise RuntimeError(
                "Both OutboundMessageFactory and "
                "InboundMessageDeserializer must be provided"
            )

        if protocol == Protocol.JSON_SINGLE:
            if outbound_message_factory is None:
                self.outbound_message_factory = JsonSingleOutboundMessageFactory(
                    device.key
                )
            else:
                if not isinstance(
                    outbound_message_factory, OutboundMessageFactory
                ):
                    raise RuntimeError(
                        "Invalid outbound message factory provided"
                    )
                else:
                    self.outbound_message_factory = outbound_message_factory

            if inbound_message_deserializer is None:
                self.inbound_message_deserializer = JsonSingleInboundMessageDeserializer(
                    self.device
                )
            else:
                if not isinstance(
                    inbound_message_deserializer, InboundMessageDeserializer
                ):
                    raise RuntimeError(
                        "Invalid inbound message deserializer provided"
                    )
                else:
                    self.inbound_message_deserializer = (
                        inbound_message_deserializer
                    )

        elif protocol == Protocol.JSON_PROTOCOL:
            if outbound_message_factory is None:
                self.outbound_message_factory = JsonProtocolOutboundMessageFactory(
                    device.key
                )
            else:
                if not isinstance(
                    outbound_message_factory, OutboundMessageFactory
                ):
                    raise RuntimeError(
                        "Invalid outbound message factory provided"
                    )
                else:
                    self.outbound_message_factory = outbound_message_factory

            if inbound_message_deserializer is None:
                self.inbound_message_deserializer = JsonProtocolInboundMessageDeserializer(
                    self.device
                )
            else:
                if not isinstance(
                    inbound_message_deserializer, InboundMessageDeserializer
                ):
                    raise RuntimeError(
                        "Invalid inbound message deserializer provided"
                    )
                else:
                    self.inbound_message_deserializer = (
                        inbound_message_deserializer
                    )
        else:
            raise RuntimeError("Unknown protocol specified - ", protocol)

        wolk_ca_cert = os.path.join(os.path.dirname(__file__), "ca.crt")

        if host and port and ca_cert:
            self.connectivity_service = OSMQTTConnectivityService(
                device,
                self.inbound_message_deserializer.inbound_topics,
                host=host,
                port=int(port),
                ca_cert=ca_cert,
            )
        elif host and port:
            self.connectivity_service = OSMQTTConnectivityService(
                device,
                self.inbound_message_deserializer.inbound_topics,
                host=host,
                port=int(port),
            )
        else:
            if connectivity_service is not None:
                if not isinstance(connectivity_service, ConnectivityService):
                    raise RuntimeError("Invalid connectivity service provided")
                else:
                    self.connectivity_service = connectivity_service
            else:
                self.connectivity_service = OSMQTTConnectivityService(
                    device,
                    self.inbound_message_deserializer.inbound_topics,
                    ca_cert=wolk_ca_cert,
                )

        self.connectivity_service.set_inbound_message_listener(
            self._on_inbound_message
        )

        if keep_alive_enabled and protocol == Protocol.JSON_SINGLE:

            self.logger.debug(
                "Keep alive enabled, interval: %s", PING_KEEP_ALIVE_INTERVAL
            )
            self.keep_alive_service = OSKeepAliveService(
                self.connectivity_service,
                self.outbound_message_factory,
                PING_KEEP_ALIVE_INTERVAL,
            )
        else:
            self.keep_alive_service = None

        self.firmware_update = OSFirmwareUpdate(firmware_handler)

        self.firmware_update.set_on_file_packet_request_callback(
            self._on_packet_request
        )
        self.firmware_update.set_on_status_callback(self._on_status)

        if device.actuator_references and (
            actuation_handler is None or actuator_status_provider is None
        ):
            raise RuntimeError(
                "Both ActuatorStatusProvider and ActuationHandler "
                "must be provided for device with actuators."
            )

        if firmware_handler:

            message = self.outbound_message_factory.make_from_firmware_version(
                firmware_handler.version
            )
            self.outbound_message_queue.put(message)

        self.firmware_update.report_result()

    def connect(self):
        """
        Connect the device to the WolkAbout IoT Platform.

        :raises e: Exception from MQTT client
        """
        try:

            self.logger.debug("connect started")
            self.connectivity_service.connect()
            if self.keep_alive_service:
                self.keep_alive_service.start()
            self.logger.debug("connect ended")

        except Exception as e:

            raise e

    def disconnect(self):
        """Disconnect the device from WolkAbout IoT Platform."""
        self.logger.debug("disconnect started")
        self.connectivity_service.disconnect()
        if self.keep_alive_service:
            self.keep_alive_service.stop()
        self.logger.debug("disconnect ended")

    def add_sensor_reading(self, reference, value, timestamp=None):
        """
        Publish a sensor reading to WolkAbout IoT Platform.

        :param reference: The reference of the sensor
        :type reference: str
        :param value: The value of the sensor reading
        :type value: int or float or str
        :param timestamp: Unix timestamp. If not provided, platform will assign
        :type timestamp: int or None
        """
        self.logger.debug("add_sensor_reading started")
        reading = SensorReading(reference, value, timestamp)
        outbound_message = self.outbound_message_factory.make_from_sensor_reading(
            reading
        )
        self.outbound_message_queue.put(outbound_message)
        self.logger.debug("add_sensor_reading ended")

    def add_alarm(self, reference, active, timestamp=None):
        """
        Publish an alarm to WolkAbout IoT Platform.

        :param reference: Reference of the alarm
        :type reference: str
        :param active: Current state of the alarm
        :type active: bool
        :param timestamp: Unix timestamp. If not provided, platform will assign
        :type timestamp: int or None
        """
        self.logger.debug("add_alarm started")
        alarm = Alarm(reference, active, timestamp)
        outbound_message = self.outbound_message_factory.make_from_alarm(alarm)
        self.outbound_message_queue.put(outbound_message)
        self.logger.debug("add_alarm ended")

    def publish(self):
        """Publish all currently stored messages to WolkAbout IoT Platform."""
        self.logger.debug("publish started")
        while True:
            outbound_message = self.outbound_message_queue.peek()

            if outbound_message is None:
                break

            if self.connectivity_service.publish(outbound_message) is True:
                self.outbound_message_queue.get()
            else:
                self.logger.warning(
                    "Failed to publish message: ",
                    outbound_message.topic,
                    outbound_message.payload,
                )
                break
        self.logger.debug("publish ended")

    def publish_actuator_status(self, reference):
        """
        Publish the current actuator status to WolkAbout IoT Platform.

        :param reference: reference of the actuator
        :type reference: str
        """
        self.logger.debug("publish_actuator_status started")
        state, value = self.actuator_status_provider.get_actuator_status(
            reference
        )
        actuator_status = ActuatorStatus(reference, state, value)
        outbound_message = self.outbound_message_factory.make_from_actuator_status(
            actuator_status
        )

        if not self.connectivity_service.publish(outbound_message):
            self.outbound_message_queue.put(outbound_message)
        self.logger.debug("publish_actuator_status ended")

    def publish_configuration(self):
        """Publish current device configuration to WolkAbout IoT Platform."""
        self.logger.debug("publish_configuration started")
        configuration = self.configuration_provider.get_configuration()
        outbound_message = self.outbound_message_factory.make_from_configuration(
            configuration
        )

        if not self.connectivity_service.publish(outbound_message):
            self.outbound_message_queue.put(outbound_message)
        self.logger.debug("publish_configuration ended")

    def _on_inbound_message(self, message):
        """
        Handle inbound messages.

        :param message: message received from the platform
        :type message: wolk.models.InboundMessage.InboundMessage
        """
        self.logger.debug("_on_inbound_message started")

        if self.inbound_message_deserializer.is_actuation_command(message):

            if not self.actuation_handler or not self.actuator_status_provider:
                return

            actuation = self.inbound_message_deserializer.deserialize_actuator_command(
                message
            )

            if actuation.command == ActuatorCommandType.SET:

                self.actuation_handler.handle_actuation(
                    actuation.reference, actuation.value
                )

                self.publish_actuator_status(actuation.reference)

            elif actuation.command == ActuatorCommandType.STATUS:

                self.publish_actuator_status(actuation.reference)

        elif self.inbound_message_deserializer.is_configuration(message):

            if (
                not self.configuration_provider
                or not self.configuration_handler
            ):
                return

            configuration = self.inbound_message_deserializer.deserialize_configuration_command(
                message
            )

            if configuration.command == ConfigurationCommandType.SET:
                self.configuration_handler.handle_configuration(
                    configuration.values
                )
                self.publish_configuration()

            elif configuration.command == ConfigurationCommandType.CURRENT:
                self.publish_configuration()

        elif self.inbound_message_deserializer.is_firmware_command(message):

            if not self.firmware_update:
                # Firmware update disabled
                firmware_status = FirmwareStatus(
                    FirmwareStatusType.ERROR,
                    FirmwareErrorType.FILE_UPLOAD_DISABLED,
                )
                outbound_message = self.outbound_message_factory.make_from_firmware_status(
                    firmware_status
                )
                if not self.connectivity_service.publish(outbound_message):
                    self.outbound_message_queue.put(outbound_message)
                return

            firmware_command = self.inbound_message_deserializer.deserialize_firmware_command(
                message
            )

            if firmware_command.command == FirmwareCommandType.FILE_UPLOAD:

                self.firmware_update.handle_file_upload(firmware_command)

            elif firmware_command.command == FirmwareCommandType.URL_DOWNLOAD:

                self.firmware_update.handle_url_download(firmware_command)

            elif firmware_command.command == FirmwareCommandType.INSTALL:

                self.firmware_update.handle_install()

            elif firmware_command.command == FirmwareCommandType.ABORT:

                self.firmware_update.handle_abort()

            elif firmware_command.command == FirmwareCommandType.UNKNOWN:
                pass

        elif self.inbound_message_deserializer.is_file_chunk(message):

            if not self.firmware_update:
                # Firmware update disabled
                firmware_status = FirmwareStatus(
                    FirmwareStatusType.ERROR,
                    FirmwareErrorType.FILE_UPLOAD_DISABLED,
                )
                outbound_message = self.outbound_message_factory.make_from_firmware_status(
                    firmware_status
                )
                if not self.connectivity_service.publish(outbound_message):
                    self.outbound_message_queue.put(outbound_message)
                return

            packet = self.inbound_message_deserializer.deserialize_firmware_chunk(
                message
            )
            self.firmware_update.handle_packet(packet)

        self.logger.debug("_on_inbound_message ended")

    def _on_packet_request(self, file_name, chunk_index, chunk_size):
        """
        Handle firmware packet requests.

        :param file_name: The name of the file to which the chunk belongs
        :type file_name: str
        :param chunk_index: The index of the requested chunk
        :type chunk_index: int
        :param chunk_size: The size of the requested chunk
        :type chunk_size: int
        """
        self.logger.debug("_on_packet_request started")
        message = self.outbound_message_factory.make_from_chunk_request(
            file_name, chunk_index, chunk_size
        )
        if not self.connectivity_service.publish(message):
            self.outbound_message_queue.put(message)
        self.logger.debug("_on_packet_request ended")

    def _on_status(self, status):
        """
        Report firmware update status to WolkAbout IoT Platform.

        :param status: The status to be reported to the platform
        :type status: wolk.models.FirmwareStatus.FirmwareStatus
        """
        self.logger.debug("_on_status started")
        message = self.outbound_message_factory.make_from_firmware_status(
            status
        )
        if not self.connectivity_service.publish(message):
            self.outbound_message_queue.put(message)
        self.logger.debug("_on_status ended")



#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""JsonProtocolInboundMessageDeserializer Module."""

import json

from wolk.models.ActuatorCommand import ActuatorCommand
from wolk.models.ActuatorCommandType import ActuatorCommandType
from wolk.models.ConfigurationCommand import ConfigurationCommand
from wolk.models.ConfigurationCommandType import ConfigurationCommandType
from wolk.models.FileTransferPacket import FileTransferPacket
from wolk.models.FirmwareCommand import FirmwareCommand
from wolk.models.FirmwareCommandType import FirmwareCommandType
from wolk.interfaces.InboundMessageDeserializer import InboundMessageDeserializer
from wolk import LoggerFactory


class JsonProtocolInboundMessageDeserializer(InboundMessageDeserializer):
    """
    Deserialize messages received from the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    """

    def __init__(self, device):
        """
        Create inbound topics from device key.

        :param device: Device with key and actuator references used for inbound topics 
        :type message: wolk.models.Device.Device
        """
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug("Init: device %s", device)

        inbound_topics = [
            "p2d/configuration_get/d/" + device.key,
            "p2d/configuration_set/d/" + device.key,
            # "p2d/file_binary_response/d/" + device.key,
            # "p2d/file_delete/d/" + device.key,
            # "p2d/file_list_confirm/d/" + device.key,
            # "p2d/file_list_request/d/" + device.key,
            # "p2d/file_purge/d/" + device.key,
            # "p2d/file_upload_abort/d/" + device.key,
            # "p2d/file_upload_initiate/d/" + device.key,
            # "p2d/file_url_download_abort/d/" + device.key,
            # "p2d/file_url_download_initiate/d/" + device.key,
            # "p2d/firmware_update_abort/d/" + device.key,
            # "p2d/firmware_update_install/d/" + device.key,
        ]

        for reference in device.actuator_references:
            inbound_topics.append(
                "p2d/actuator_set/d/" + device.key + "/r/" + reference
            )
            inbound_topics.append(
                "p2d/actuator_get/d/" + device.key + "/r/" + reference
            )

        super().__init__(inbound_topics)

    def is_actuation_command(self, message):
        """
        Check if message is actuation command

        :param message: The message received
        :type message: wolk.models.InboundMessage.InboundMessage
        :returns: actuation_command
        :rtype: bool
        """
        if message.topic.startswith("p2d/actuator"):
            return True
        return False

    def is_firmware_command(self, message):
        """
        Check if message is firmware command

        :param message: The message received
        :type message: wolk.models.InboundMessage.InboundMessage
        :returns: firmware_command
        :rtype: bool
        """
        pass

    def is_file_chunk(self, message):
        """
        Check if message is file chunk

        :param message: The message received
        :type message: wolk.models.InboundMessage.InboundMessage
        :returns: file_chunk
        :rtype: bool
        """
        pass

    def is_configuration(self, message):
        """
        Check if message is configuration

        :param message: The message received
        :type message: wolk.models.InboundMessage.InboundMessage
        :returns: configuration
        :rtype: bool
        """
        if message.topic.startswith("p2d/configuration"):
            return True
        return False

    def deserialize_actuator_command(self, message):
        """
        Deserialize the message into an actuation command.

        :param message: Message to be deserialized
        :type message: wolk.models.InboundMessage.InboundMessage

        :returns: actuation
        :rtype: wolk.models.ActuatorCommand.ActuatorCommand
        """
        self.logger.debug("deserialize_actuator_command called")
        reference = message.topic.split("/")[-1]
        payload = json.loads(message.payload.decode("utf-8"))

        if "actuator_set" in message.topic:

            command_type = ActuatorCommandType.SET
            value = payload.get("value")
            if "\n" in value:
                value = value.replace("\n", "\\n")
                value = value.replace("\r", "")
            if value == "true":
                value = True
            elif value == "false":
                value = False

            actuation = ActuatorCommand(reference, command_type, value)
            self.logger.info(
                "Received actuation command - Reference: %s ; Command: SET ;"
                " Value: %s",
                actuation.reference,
                actuation.value,
            )
            return actuation

        elif "actuator_get" in message.topic:

            command_type = ActuatorCommandType.STATUS
            actuation = ActuatorCommand(reference, command_type)
            self.logger.info(
                "Received actuation command - Reference: %s ;" " Command: STATUS ",
                actuation.reference,
            )
            return actuation

        else:

            command_type = ActuatorCommandType.UNKNOWN
            actuation = ActuatorCommand(reference, command_type)
            self.logger.warning(
                "Received unknown actuation command on topic - : %s ; Payload: %s",
                message.topic,
                message.payload,
            )
            return actuation

    def deserialize_firmware_command(self, message):
        """
        Deserialize the message into a firmware command.

        :param message: Message to be deserialized
        :type message: wolk.models.InboundMessage.InboundMessage

        :returns: firmware_command
        :rtype: wolk.models.FirmwareCommand.FirmwareCommand
        """
        pass

    def deserialize_firmware_chunk(self, message):
        """
        Split the message into a packet.

        :param message: Message to be deserialized
        :type message: wolk.models.InboundMessage.InboundMessage

        :returns: packet
        :rtype: wolk.models.FileTransferPacket.FileTransferPacket
        """
        pass

    def deserialize_configuration_command(self, message):
        """
        Deserialize the message into a configuration command.

        :param message: message to be deserialized
        :type message: wolk.models.InboundMessage.InboundMessage

        :returns: configuration
        :rtype: wolk.models.ConfigurationCommand.ConfigurationCommand
        """
        self.logger.debug("deserialize_configuration_command called")
        payload = json.loads(message.payload.decode("utf-8"))

        if "configuration_set" in message.topic:

            command = ConfigurationCommandType.SET

            configuration = ConfigurationCommand(command, payload)

            self.logger.info(
                "Received configuration command - Command: SET  Values: %s",
                configuration.values,
            )

            for reference, value in configuration.values.items():
                if "\n" in value:
                    value = value.replace("\n", "\\n")
                    value = value.replace("\r", "")
                if value == "true":
                    value = True
                elif value == "false":
                    value = False

                if isinstance(value, bool):
                    pass
                else:
                    if "," in value:
                        values_list = value.split(",")
                        try:
                            if any("." in value for value in values_list):
                                values_list = [float(value) for value in values_list]
                            else:
                                values_list = [int(value) for value in values_list]
                        except ValueError:
                            pass

                        configuration.values[reference] = tuple(values_list)

            return configuration

        elif "configuration_get" in message.topic:

            command = ConfigurationCommandType.CURRENT

            configuration = ConfigurationCommand(command)
            self.logger.info("Received configuration command - Command: CURRENT")
            return configuration

        else:

            command = ConfigurationCommandType.UNKNOWN

            configuration = ConfigurationCommand(command)
            self.logger.warning(
                "Received configuration command - Command: %s", configuration.command
            )
            return configuration



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
FirmwareCommand Module.
"""


class FirmwareCommand:
    """Firmware update command received from WolkAbout IoT Platform."""

    def __init__(
        self,
        command,
        file_name=None,
        file_size=None,
        file_hash=None,
        auto_install=None,
        file_url=None,
    ):
        """
        Firmware command that was received.

        :param command: Command to be executed
        :type command: wolk.models.FirmwareCommandType.FirmwareCommandType
        :param file_name: Name of the file to be received
        :type file_name: str or None
        :param file_size: Size of the file to be received
        :type file_size: int or None
        :param file_hash: Hash of the file to be received
        :type file_hash: str or None
        :param auto_install: Install the file when it is received
        :type auto_install: bool or None
        :param file_url: URL where the file is located
        :type file_url: str or None
        """
        self.command = command
        self.file_name = file_name
        self.file_size = file_size
        self.file_hash = file_hash
        self.auto_install = auto_install
        self.file_url = file_url



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

"""
FirmwareHandler Module.
"""


class FirmwareHandler(ABC):
    """
    Handle firmware update process.

    Implement this class to write and read files in persistent memory,
    persist firmware version and handle the process
    of installation of firmware file itself.
    Optionally provide implementation for URL download
    """

    @abstractmethod
    def update_start(self, file_name, file_size):
        """
        Start receiving the firmware file.

        :param file_name: name of the firmware file
        :type file_name: str
        :param file_size: size of the firmware file in bytes
        :type file_size: int

        :returns: result
        :rtype: bool
        """
        pass

    @abstractmethod
    def update_finalize(self):
        """Finish the process of firmware update."""
        pass

    @abstractmethod
    def update_abort(self):
        """Abort the process of firmware update."""
        pass

    @abstractmethod
    def write_chunk(self, chunk):
        """
        Write a chunk of the firmware file to the temporary file.

        :param chunk: A piece of the firmware file
        :type chunk: bytes

        :returns: result
        :rtype: bool
        """
        pass

    @abstractmethod
    def read_chunk(self, index):
        """
        Read a chunk of the temporary firmware file.

        :param index: offset from the beginning of the file
        :type index: int

        :returns: chunk
        :rtype: bytes
        """
        pass

    @abstractmethod
    def persist_version(self, version):
        """
        Place the current firmware version into persistent storage.

        Later to be used to determine the result
        of the firmware update process

        :param version: The current firmware version
        :type version: str

        :returns: result
        :rtype: bool
        """
        pass

    @abstractmethod
    def unpersist_version(self):
        """
        Remove the firmware version from persistent storage.

        :returns: version
        :rtype: str or None
        """
        pass

    @abstractmethod
    def update_start_url_download(self, file_url):
        """
        Prepare the device for the URL file transfer.

        :param file_url: The URL that contains the firmware file
        :type file_url: str

        :returns: result
        :rtype: bool
        """
        pass

    @abstractmethod
    def set_url_download_result_callback(self, callback):
        """
        Set the callback function for reporting the result of URL file download.

        :param callback: function to call
        :type callback: function
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

"""
OutboundMessageQueue Module.
"""


class OutboundMessageQueue(ABC):
    """Store messages on device before publishing to WolkAbout IoT Platform."""

    @abstractmethod
    def put(self, message):
        """
        Place a message in storage.

        :param message: Message to be stored
        :type message: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def get(self):
        """
        Get a message from storage.

        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass

    @abstractmethod
    def peek(self):
        """
        Get a message without removing from storage.

        :returns: message
        :rtype: wolk.models.OutboundMessage.OutboundMessage
        """
        pass
